/**
 * Assignment #6: Page Replacement Algorithms
 * Sample Solution
 *
 * Instructor: Ron Mak
 * CMPE 142 Operating Systems
 * Computer Engineering Department
 * San Jose State University
 */
#include <iostream>
#include <vector>
#include <stdlib.h>

#include "ReplacementAlgorithm.h"
#include "FIFO.h"
#include "LRU.h"
#include "LFU.h"
#include "MFU.h"
#include "Random.h"
#include "Optimal.h"

//#define DEBUG

using namespace std;

const int FRAME_COUNT      =   3;
const int MAX_REFERENCE    =   7;
const int STRING_LENGTH    = 100;
const int SIMULATION_COUNT =   5;

void generate_reference_string(vector<int>& references,
                               const int max_ref, const int string_length);

void do_simulations(const int sim_count, const int max_ref,
                    const int string_length);

int main()
{
    srand(time(0));
    do_simulations(SIMULATION_COUNT, MAX_REFERENCE, STRING_LENGTH);

    return 0;
}

void do_simulations(const int sim_count, const int max_ref,
                    const int string_length)
{
    vector<int> reference_string;
    vector<int> frames;

    vector<ReplacementAlgorithm *> algorithms
    {
        new FIFO("FIFO", reference_string, MAX_REFERENCE, frames, FRAME_COUNT),
        new LRU("LRU", reference_string, MAX_REFERENCE, frames, FRAME_COUNT),
        new LFU("LFU", reference_string, MAX_REFERENCE, frames, FRAME_COUNT),
        new MFU("MFU", reference_string, MAX_REFERENCE, frames, FRAME_COUNT),
        new Random("Random", reference_string, MAX_REFERENCE, frames, FRAME_COUNT),
        new Optimal("Optimal", reference_string, MAX_REFERENCE, frames, FRAME_COUNT),
    };

    for (int i = 1; i <= sim_count; i++)
    {
        cout << endl;
        cout << "=================" << endl;
        cout << "SIMULATION RUN #"  << i << endl;
        cout << "=================" << endl;

        generate_reference_string(reference_string, max_ref, string_length);

        cout << endl << "Reference string:" << endl;
        for (int ref : reference_string) cout << ref;
        cout << endl;

        for (auto alg : algorithms) alg->run(i);
    }

    cout << endl;
    cout << "==================" << endl;
    cout << "SUMMARY STATISTICS" << endl;
    cout << "==================" << endl << endl;

    int ref_counts = sim_count*reference_string.size();

    cout << "         Number of simulations = " << sim_count << endl;
    cout << "       Reference string length = " << reference_string.size() << endl;
    cout << "Total references per algorithm = " << ref_counts << endl << endl;

    cout << "Algorithm   Total swaps   Total hits   Overall hit ratio" << endl;

    for (auto alg : algorithms)
    {
        const char *name       = alg->get_name().c_str();
        double      swap_count = alg->get_total_swap_count();
        int         hit_count  = alg->get_total_hit_count();
        double      hit_ratio  = (double)alg->get_total_hit_count() / ref_counts;

        printf("%8s%13.2f%12d%17.2f\n",
               name, swap_count, hit_count, hit_ratio);
    }
}

#ifdef DEBUG
const int TEST_STRING[] = {7,0,1,2,0,3,0,4,2,3,0,3,2,1,2,0,1,7,0,1,-1};
#endif

void generate_reference_string(vector<int>& ref_string,
                               const int max_ref, const int string_length)
{
    ref_string.clear();

#ifdef DEBUG
    int i = 0;

    do
    {
        ref_string.push_back(TEST_STRING[i]);
    } while (TEST_STRING[++i] >= 0);
#else
    int ref = rand()%max_ref;

    while (ref_string.size() < string_length)
    {
        ref_string.push_back(ref);

        switch (rand()%10)
        {
            case 0:
            case 1:
            {
                ref--;
                if (ref < 0) ref = max_ref;
                break;
            }

            case 2:
            case 3: break;

            case 4:
            case 5:
            {
                ref++;
                if (ref > max_ref) ref = 0;
                break;
            }

            default:
            {
                ref = (ref + (rand()%max_ref + 1))%max_ref;
                break;
            }
        }
    }
#endif
}
